/*******************************************************************************
 * rexxplug.c
*******************************************************************************/

#define INCL_REXXSAA
 
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <rexxsaa.h>
#include "../include/npapi.h"
#include "../include/npupp.h"

#ifdef XP_PC
#include <io.h>
#endif


/*******************************************************************************
 * size of streambuffer for stream-based plugin
*******************************************************************************/
#define STREAMBUFSIZE 4096

/*******************************************************************************
 * instance data definition
*******************************************************************************/
typedef struct _PluginInstance
{
	NPWindow	*Window;				/* handle of the plugin window */
	int			tmpFile;				/* handle of temporary scripting file */
	char 		tmpFileName[L_tmpnam];	/* L_tmpnam <stdio.h>, max. size of temp. filename */
	char		*target;				/* target for the output stream or URL */
	char		*input; 				/* input for rexx stdio */
	NPStream	*rexx_stream;		    /* pointer to the stream created by the plugin */

#ifdef XP_PC
	int			fHTML;					/* Handle fr temp. Skriptdatei */
	char 		fHTMLFileName[L_tmpnam];/* L_tmpnam aus <stdio.h>, max. lnge temp. Dateiname */	
#endif
} PluginInstance;

/*******************************************************************************
 * function prototypes
*******************************************************************************/
NPError rexx_exec(NPP instance);

/*******************************************************************************
 * UNIX-Only API calls
 ******************************************************************************/
#ifdef XP_UNIX
char*
NPP_GetMIMEDescription(void)
{
	return("application/x-rexx:rex:Rexx Plugin");
}

NPError
NPP_GetValue(void *future, NPPVariable variable, void *value)
{
	NPError err = NPERR_NO_ERROR;

	switch (variable) {
		case NPPVpluginNameString:			// Browser asks for the plugin name
			*((char **)value) = "Rexx plugin 0.1";
			break;
		case NPPVpluginDescriptionString:	// Browser asks for a description of the plugin
			*((char **)value) =
				"Rexx Scripting Support";
			break;
		default:
			err = NPERR_GENERIC_ERROR;
	}
	return err;
}                                                     
#endif /* XP_UNIX */

/*******************************************************************************
 * plugin initializing
 *
 * called, before the first plugin instance is created
 ******************************************************************************/
NPError
NPP_Initialize(void)
{
	return NPERR_NO_ERROR;
}

/*******************************************************************************
 * plugin shutdown
 *
 * called, after the last plugin instance is destroyed
 ******************************************************************************/
void
NPP_Shutdown(void)
{
	/* nothing to do... */
}

/*******************************************************************************
 * creating an instance of the plugin
 *
 * by clicking to a link with an EMBED-Tag that specifies this plugin,
 * the browser passes a pointer to an instance structur
 ******************************************************************************/
void 
copy_param(char** dest, const char* src)
{
	if (src != NULL) {	
		*dest = NPN_MemAlloc(strlen(src) + 1);
		memset(*dest, 0, strlen(src) + 1);
		memcpy(*dest, src, strlen(src));		
	} 
	else {
		dest = NULL;
	}	
}

NPError 
NPP_New(NPMIMEType pluginType,
	NPP instance,
	uint16 mode,
	int16 argc,
	char* argn[],
	char* argv[],
	NPSavedData* saved)
{
	int i;
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	/* 	We have to allocate memory for the instance data.
		For the lifetime of this plugin instance it's our
		private instance data.
		The Browsers private data is stored in 
		instance->ndata, but we must not modify it! */		
	instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
	memset(instance->pdata, 0, sizeof(PluginInstance));

	This = (PluginInstance*) instance->pdata;
	if (This == NULL)
		return NPERR_OUT_OF_MEMORY_ERROR;		
	
	/* examine the parameter list */
	for (i=0; i<argc; i++) {
		if (strcmp(argn[i], "target") == 0)
			copy_param(&This->target, argv[i]);
		if (strcmp(argn[i], "input") == 0)
			copy_param(&This->input, argv[i]);
	}
	
	/* set default parameters */
	if (This->target == NULL)
			copy_param(&This->target, "_self");
						
	/* Create a temporary filename, where we can save the rexx-source */
	tmpnam(This->tmpFileName);
	if ((This->tmpFile = creat(This->tmpFileName, S_IWRITE | S_IREAD)) < 0)
		return NPERR_GENERIC_ERROR;

#ifdef XP_UNIX
	This->rexx_stream = NULL;
#endif

#ifdef XP_PC
	This->fHTML = 0;		
#endif
	return NPERR_NO_ERROR;
}

/*******************************************************************************
 * destroing a plugin instance
 ******************************************************************************/
NPError 
NPP_Destroy(NPP instance, NPSavedData** save)
{
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	/* 	get a reference to our private instance data */
	This = (PluginInstance*) instance->pdata;

	if (This != NULL) {	
		/* delete the temporary file, which was created in NPP_New */
		unlink(This->tmpFileName);	
		
		/* Free the allocated memory of our private instance data */
		NPN_MemFree(This->target);
		NPN_MemFree(This->input);
		NPN_MemFree(instance->pdata);
		instance->pdata = NULL;
	}

	return NPERR_NO_ERROR;
}

/*******************************************************************************
 * initiation of streaming
 ******************************************************************************/
NPError 
NPP_NewStream(NPP instance,
	      NPMIMEType type,
	      NPStream *stream, 
	      NPBool seekable,
	      uint16 *stype)
{
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;
	
	return NPERR_NO_ERROR;
}

/*******************************************************************************
 *  set the amount of data per NPP_Write-Call in bytes
 ******************************************************************************/
int32 
NPP_WriteReady(NPP instance, NPStream *stream)
{
	PluginInstance* This;
	
	if (instance != NULL)
		This = (PluginInstance*) instance->pdata;

	return STREAMBUFSIZE;
}


/*******************************************************************************
 *  Receiving Data from the Webserver
 ******************************************************************************/
int32 
NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
	if (instance != NULL)
	{
		PluginInstance* This = (PluginInstance*) instance->pdata;
		
		return write(This->tmpFile, (char*)buffer, (size_t)(len<=STREAMBUFSIZE ? len : STREAMBUFSIZE));
	}

	return -1;
}


/*******************************************************************************
 *  data receive complete, destroy the stream
 ******************************************************************************/
NPError 
NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
	PluginInstance* This;
	
	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;

	if (stream != This->rexx_stream) {
		close(This->tmpFile);
		return rexx_exec(instance);
	}
	
	return NPERR_NO_ERROR;
}

/*******************************************************************************
 *  received data is stored by the browser to a local file
 ******************************************************************************/
void 
NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
	if (instance != NULL) {
		PluginInstance* This; 
		This = (PluginInstance*) instance->pdata;
	} 
}

/*******************************************************************************
 *  Hook for the output of REXX-command SAY
 ******************************************************************************/
#ifdef XP_UNIX
APIRET handle_RXSIOSAY(PEXIT ParmBlock) {
	NPP				instance;
	PluginInstance* This;
    void**			UserDataAddr;
	USHORT			Flag;

    UserDataAddr = NPN_MemAlloc(8);
    
    RexxQueryExit("hook_RXSIOSAY", NULL, &Flag, (PUCHAR)UserDataAddr);

	instance = (NPP)*UserDataAddr;
	This = (PluginInstance*)instance->pdata;
	
    if (This->rexx_stream == NULL) {
		if (NPN_NewStream(instance, "text/html", This->target, &This->rexx_stream) != NPERR_NO_ERROR) {
			NPN_MemFree(UserDataAddr);
			return RXEXIT_HANDLED;
		}
	}
	NPN_Write(	instance, 
				This->rexx_stream, 
				((EXIT*)ParmBlock)->siosay.rxsio_string.strlength, 
				((EXIT*)ParmBlock)->siosay.rxsio_string.strptr);

	NPN_MemFree(UserDataAddr);

	return RXEXIT_HANDLED;
}
#endif /* XP_ PC */

/******************************************************************************
 *  Because it's not possible to initiate a stream in Internet Explorer,
 *	the output of SAY is saved to a temporary HTML-file. Then this file is 
 *	to be viewed in a further Internet Explorer window
 ******************************************************************************/
#ifdef XP_PC
APIRET handle_RXSIOSAY(PEXIT ParmBlock) {
	NPP				instance;
	PluginInstance* This;
    void**			UserDataAddr;
	USHORT			Flag;
	char*			tmp;

    UserDataAddr = NPN_MemAlloc(8);
    RexxQueryExit("hook_RXSIOSAY", NULL, &Flag, (PUCHAR)UserDataAddr);

	instance = (NPP)*UserDataAddr;
	This = (PluginInstance*)instance->pdata;

	if (This->fHTML == 0) {
		tmp = _tempnam(NULL, "rexx");		
		strcpy(This->fHTMLFileName, tmp);
		free(tmp);
		if ((This->fHTML = creat(This->fHTMLFileName, S_IWRITE | S_IREAD)) < 0) {
			NPN_MemFree(UserDataAddr);		
			return RXEXIT_HANDLED;
		}
    }
    
	write(	This->fHTML, 
			((EXIT*)ParmBlock)->siosay.rxsio_string.strptr, 
			((EXIT*)ParmBlock)->siosay.rxsio_string.strlength);

	NPN_MemFree(UserDataAddr);

	return RXEXIT_HANDLED;
}
#endif /* XP_ PC */

/*******************************************************************************
 * REXX Exit-Handler
 ******************************************************************************/
APIRET APIENTRY hook_RXSIOSAY(LONG ExitNumber, LONG Subfunction, PEXIT ParmBlock) {
	switch (ExitNumber) {
		case RXSIO:
			switch (Subfunction) {
				case RXSIOSAY: return handle_RXSIOSAY(ParmBlock);
				default: return RXEXIT_NOT_HANDLED;
			}
		default: return RXEXIT_NOT_HANDLED;
	}
}

/*******************************************************************************
 * REXX execute rexx interpreter
 ******************************************************************************/
NPError
rexx_exec(NPP instance) 
{  
	PluginInstance* This;
	long			UserDataAddr;
    SHORT			ReturnCode;
    RXSTRING		Result;
    RXSYSEXIT		sysexit[2];

	RXSTRING		arglist[1];
#ifdef XP_PC
	char 			url[L_tmpnam];
#endif /* XP_PC */

	This = (PluginInstance*) instance->pdata;
	UserDataAddr = (long)instance;

    sysexit[0].sysexit_code = RXSIO;
    sysexit[0].sysexit_name = "hook_RXSIOSAY";
    sysexit[1].sysexit_code = RXENDLST;
		
	arglist[0].strptr = This->input;
	arglist[0].strlength = strlen(This->input);

	RexxRegisterExitExe("hook_RXSIOSAY", hook_RXSIOSAY, (PUCHAR)&UserDataAddr);
	RexxStart(1, arglist, This->tmpFileName, NULL, NULL, RXCOMMAND, sysexit, &ReturnCode, &Result);
	RexxDeregisterExit("hook_RXSIOSAY", NULL);

#ifdef XP_PC
	if (This->fHTML != 0) {			
		close(This->fHTML);
		sprintf(url, "file://%s", This->fHTMLFileName);
		if (!NPN_GetURL((NPP)instance, url, This->target) == NPERR_NO_ERROR)
			return NPERR_GENERIC_ERROR;
	}
#endif /* XP_PC */

#ifdef XP_UNIX
	if (This->rexx_stream != NULL)
	 	NPN_DestroyStream(instance, This->rexx_stream, NPRES_DONE);			
#endif /* XP_UNIX */
			
	return NPERR_NO_ERROR;
}

/*******************************************************************************
 * Functions, that must be implemented in the plugin but we don't need!
 ******************************************************************************/

/*******************************************************************************
 * refresh the plugin window
 ******************************************************************************/
NPError 
NPP_SetWindow(NPP instance, NPWindow* window)
{
	/* we don't produce any output */
	return NPERR_NO_ERROR;
}

/*******************************************************************************
 *  Notifiy?
 ******************************************************************************/
void
NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData)
{
	/* Not used in this simple plugin. */
}

/*******************************************************************************
 *  Java?
 ******************************************************************************/
jref
NPP_GetJavaClass()
{
	/* we don't use java */
    return NULL;
}
/*******************************************************************************
 *  Printing ?
 ******************************************************************************/
void
NPP_Print(NPP instance, NPPrint* printInfo)
{
	/* we don't want to print */
}
